/*
   +----------------------------------------------------------------------+
   | PHP Version 4                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2003 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 2.02 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available at through the world-wide-web at                           |
   | http://www.php.net/license/2_02.txt.                                 |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Authors: Justin Hannus <justin@128kb.com>                            |
   |                                                                      |
   +----------------------------------------------------------------------+
*/
 
/* $Id: directoryservices.c,v 0.2.3 2004/05/08 09:52:01 jhannus jhannus@128kb.com Exp $ */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif


#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <DirectoryService/DirServices.h>
#include <DirectoryService/DirServicesUtils.h>
#include <DirectoryService/DirServicesConst.h>
#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_opendirectory.h"
#include "opendirectory_errors.h"


#define DS_RSRC_NAME_DIRREF    "directory reference"
#define DS_RSRC_NAME_DLIST     "data list"
#define DS_RSRC_NAME_NODEREF   "node reference"
#define DS_RSRC_NAME_RECREF    "record reference"

#define DS_KEY_REC_ATTRCOUNT   "fRecordAttributeCount"
#define DS_KEY_REC_NAMETYPE    "fRecordNameAndType"
#define DS_KEY_REC_NAME        "fRecordName"
#define DS_KEY_REC_TYPE        "fRecordType"
#define DS_KEY_ATTRVAL         "fAttributeValueData"
#define DS_KEY_ATTRVAL_ID      "fAttributeValueID"
#define DS_KEY_ATTRVAL_COUNT   "fAttributeValueCount"
#define DS_KEY_ATTRVAL_MAXSIZE "fAttributeValueMaxSize"
#define DS_KEY_ATTR_SIZE       "fAttributeDataSize"
#define DS_KEY_ATTR_SIG        "fAttributeSignature"

#define PHP_DS_DLIST_GET       0
#define PHP_DS_DLIST_FIND      1

#define DS_BUFFTYPE_NONE       0
#define DS_BUFFTYPE_NODELIST   1
#define DS_BUFFTYPE_RECLIST    2

#define REMOTE_SESSION         0
#define LOCAL_SESSION          1

#define PHP_DS_STATUS(__s) DS_G(status) = __s;
#define PHP_DS_STATUS_RETURN_FALSE(s) {                    \
                                         PHP_DS_STATUS(s); \
                                         RETURN_FALSE;     \
                                      }
                                        
#define PHP_DS_STATUS_RETURN_TRUE(s) {                     \
                                        PHP_DS_STATUS(s);  \
                                        RETURN_TRUE;       \
                                     }

#define PHP_DS_STATUS_RETURN_NULL(s) {                     \
                                        PHP_DS_STATUS(s);  \
                                        RETURN_NULL();     \
                                     }
                                        
#define PHP_DBUFFER_P_INIT(dref) {                                                                        \
                             dref->data = (php_t_data_buffer *) emalloc(sizeof(php_t_data_buffer));       \
                             dref->data->buffer  = NULL;                                                  \
                             dref->data->context = NULL;                                                  \
                             dref->data->count   = 0;                                                     \
                        }

#define PHP_DBUFFER_DEALLOC(dref) {                                                                       \
                            if (dref->data) {                                                             \
                                if (dref->data->buffer) {                                                 \
                                    PHP_DS_STATUS(dsDataBufferDeAllocate(dref->ref, dref->data->buffer)); \
                                    dref->data->buffer = NULL;                                            \
                                }                                                                         \
                                if (dref->data->context) {                                                \
                                    PHP_DS_STATUS(dsReleaseContinueData(dref->ref, dref->data->context)); \
                                    dref->data->context = NULL;                                           \
                                }                                                                         \
                                efree(dref->data);                                                        \
                                dref->data = NULL;                                                        \
                            }                                                                             \
                        }
                        
#define PHP_DS_RESULT_INIT(res, type) {                                                                   \
                            res = (php_ds_result *) emalloc(sizeof(php_ds_result));                       \
                            res.type = type;                                                              \                                                   \
                        }
                        
                        

/* {{{
   }}}  zend list entry id's */
static int le_dirref, le_node_ref, le_rec_ref;


/* {{{
   }}}  resource types */
typedef struct _php_t_data_buffer {
    tDataBufferPtr  buffer;
    tContextData    context;
    unsigned long   count;
} php_t_data_buffer;

typedef struct _php_ds_result {
    unsigned char      type;
    php_t_data_buffer  *data;
} php_ds_result;

typedef struct _php_t_data_list {
    tDataListPtr   list;
    unsigned long  count;
} php_t_data_list;

typedef struct _php_t_dir_reference {
    tDirReference     ref;
    php_t_data_buffer *data;
    long              status;
} php_t_dir_reference;

typedef struct _php_t_dir_node_ref {
    tDirNodeReference ref;
    long              assoc_dirref_id;
} php_t_dir_node_ref;

typedef struct _php_t_rec_ref {
    tRecordReference ref;
    long             assoc_dirref_id;
} php_t_rec_ref;

typedef struct _php_node_auth_buffer {
    tDataNodePtr   type;
    tDataBufferPtr data;
    tDataBufferPtr result; 
    tContextData   context;
} php_node_auth_buffer;


/* {{{
   }}}  private function prototypes */
static void _php_le_dirref_close(zend_rsrc_list_entry   *rsrc TSRMLS_DC);
static void _php_le_node_ref_close(zend_rsrc_list_entry *rsrc TSRMLS_DC);
static void _php_le_rec_ref_close(zend_rsrc_list_entry  *rsrc TSRMLS_DC);
static void php_ds_init_globals(zend_opendirectory_globals *ds_globals);
static tDirStatus php_node_list_to_zarray(php_t_dir_reference *, php_t_data_buffer *, zval * TSRMLS_DC);
static tDirStatus php_rec_list_to_zarray(php_t_dir_reference  *, php_t_data_buffer *, php_t_dir_node_ref *, zval *, unsigned long TSRMLS_DC);
static tDirStatus php_zarray_to_data_list(php_t_dir_reference *, zval *, tDataListPtr TSRMLS_DC);
static tDirStatus php_attr_list_to_zarray(php_t_dir_reference *, php_t_dir_node_ref *, tAttributeListRef *, tAttributeValueListRef *, zval *, long, bool TSRMLS_DC);
static php_node_auth_buffer * php_alloc_node_auth_buffer(php_t_dir_reference *, char *, char *, int, char * TSRMLS_DC);
static void php_dealloc_node_auth_buffer(php_t_dir_reference *, php_node_auth_buffer * TSRMLS_DC);
static void php_ds_open_directory_service(INTERNAL_FUNCTION_PARAMETERS, int);


#ifdef ZEND_ENGINE_2
#define OnUpdateInt  OnUpdateLong
static ZEND_BEGIN_ARG_INFO(first_arg_force_ref, 0)
	ZEND_ARG_PASS_INFO(1)
ZEND_END_ARG_INFO();
#else
static unsigned char first_arg_force_ref[] = { 
	1, BYREF_FORCE 
};
#endif


/* {{{ opendirectory_functions[]
 */
zend_function_entry opendirectory_functions[] =
{
    ZEND_FE(ds_open_dir_service,               NULL)
    ZEND_FE(ds_open_dir_service_proxy,         NULL)
    ZEND_FE(ds_set_separator,                  NULL)
    ZEND_FE(ds_get_separator,                  NULL)
    ZEND_FE(ds_close_dir_service,              NULL)
    ZEND_FE(ds_verify_dir_ref_num,             NULL)
    ZEND_FE(ds_add_child_pid_to_reference,     NULL)
    ZEND_FE(ds_get_dir_node_count,             NULL)
    ZEND_FE(ds_get_dir_node_list,              NULL)
    ZEND_FE(ds_dir_status,                     NULL)
    ZEND_FE(ds_dir_errstr,                     NULL)
    ZEND_FE(ds_is_dir_service_running,         NULL)
    ZEND_FE(ds_data_buffer_alloc,              NULL)
    ZEND_FE(ds_data_buffer_dealloc,            NULL)
    ZEND_FE(ds_do_dir_node_auth,               NULL)
    ZEND_FE(ds_find_dir_nodes,                 NULL)
    ZEND_FE(ds_open_dir_node,                  NULL)
    ZEND_FE(ds_close_dir_node,                 NULL)
    ZEND_FE(ds_get_record_list,                NULL)
    ZEND_FE(ds_continue_data,                  NULL)
    ZEND_FE(ds_release_continue_data,          NULL)
    ZEND_FE(ds_close_record,                   NULL)
    ZEND_FE(ds_open_record,                    NULL)
    ZEND_FE(ds_set_record_name,                NULL)
    ZEND_FE(ds_set_record_type,                NULL)
    ZEND_FE(ds_create_record,                  NULL)
    ZEND_FE(ds_flush_record,                   NULL)
    ZEND_FE(ds_create_record_and_open,         NULL)
    ZEND_FE(ds_add_attribute,                  NULL)
    ZEND_FE(ds_add_attribute_value,            NULL)
    ZEND_FE(ds_remove_attribute,               NULL)
    ZEND_FE(ds_remove_attribute_value,         NULL)
    ZEND_FE(ds_get_record_attr_value_by_index, NULL)
    ZEND_FE(ds_get_record_attr_value_by_id,    NULL)
    ZEND_FE(ds_delete_record,                  first_arg_force_ref)
    ZEND_FE(ds_get_dir_node_info,              NULL)
    ZEND_FE(ds_get_record_reference_info,      NULL)
    ZEND_FE(ds_get_record_attribute_info,      NULL)
    ZEND_FE(ds_do_attribute_value_search,      NULL)
    {
        NULL, 
        NULL, 
        NULL
    }
};
/* }}} */


/* {{{ opendirectory_module_entry
 */
zend_module_entry opendirectory_module_entry =
{
    STANDARD_MODULE_HEADER,
    "directoryservices",
    opendirectory_functions,
    ZEND_MINIT(opendirectory),
    ZEND_MSHUTDOWN(opendirectory), 
    ZEND_RINIT(opendirectory),
    ZEND_RSHUTDOWN(opendirectory), 
    ZEND_MINFO(opendirectory),
    OPENDIRECTORY_API_VERSION,
    STANDARD_MODULE_PROPERTIES
};
/* }}} */


ZEND_DECLARE_MODULE_GLOBALS(opendirectory)

static void php_ds_init_globals(zend_opendirectory_globals *ds_globals)
{
    ds_globals->separator      = NULL;
    ds_globals->status         = eDSNoErr;
    ds_globals->num_refs       = 0;
    ds_globals->max_refs       = -1;
}


/* {{{ PHP_INI */
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("opendirectory.separator",  "/",          PHP_INI_ALL,    OnUpdateString, separator,        zend_opendirectory_globals, opendirectory_globals)
    STD_PHP_INI_ENTRY("opendirectory.max_refs",  "-1",          PHP_INI_ALL,    OnUpdateInt,    max_refs,         zend_opendirectory_globals, opendirectory_globals)
PHP_INI_END()
/* }}} */


/* {{{ ZEND_MINIT_FUNCTION
 */
ZEND_MINIT_FUNCTION(opendirectory)
{
    le_dirref   = zend_register_list_destructors_ex(_php_le_dirref_close,   NULL, DS_RSRC_NAME_DIRREF,  module_number);
    le_node_ref = zend_register_list_destructors_ex(_php_le_node_ref_close, NULL, DS_RSRC_NAME_NODEREF, module_number);
    le_rec_ref  = zend_register_list_destructors_ex(_php_le_rec_ref_close,  NULL, DS_RSRC_NAME_RECREF,  module_number);
    
    ZEND_INIT_MODULE_GLOBALS(opendirectory, php_ds_init_globals, NULL);
    REGISTER_INI_ENTRIES();
    Z_TYPE(opendirectory_module_entry) = type;

    REGISTER_LONG_CONSTANT("DS_NO_ERROR",          eDSNoErr,             CONST_PERSISTENT | CONST_CS);
    REGISTER_LONG_CONSTANT("DS_M_STARTS_WITH",     eDSStartsWith,        CONST_PERSISTENT | CONST_CS);
    REGISTER_LONG_CONSTANT("DS_M_ENDS_WITH",       eDSEndsWith,          CONST_PERSISTENT | CONST_CS);
    REGISTER_LONG_CONSTANT("DS_M_EXACT",           eDSExact,             CONST_PERSISTENT | CONST_CS);
    REGISTER_LONG_CONSTANT("DS_M_CONTAINS",        eDSContains,          CONST_PERSISTENT | CONST_CS);
    REGISTER_LONG_CONSTANT("DS_MI_STARTS_WITH",    eDSiStartsWith,       CONST_PERSISTENT | CONST_CS);
    REGISTER_LONG_CONSTANT("DS_MI_ENDS_WITH",      eDSiEndsWith,         CONST_PERSISTENT | CONST_CS);
    REGISTER_LONG_CONSTANT("DS_MI_EXACT",          eDSiExact,            CONST_PERSISTENT | CONST_CS);
    REGISTER_LONG_CONSTANT("DS_MI_CONTAINS",       eDSiContains,         CONST_PERSISTENT | CONST_CS);
    
    REGISTER_STRING_CONSTANT("DS_RECORDS_ALL",               kDSRecordsAll,             CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_RECORDS_STANDARD_ALL",      kDSRecordsStandardAll,     CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_RECORDS_NATIVE_ALL",        kDSRecordsNativeAll,       CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_ATTRIBUTES_ALL",            kDSAttributesAll,          CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_ATTRIBUTES_STANDARD_ALL",   kDSAttributesStandardAll,  CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_ATTRIBUTES_NATIVE_ALL",     kDSAttributesNativeAll,    CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_STD_RECORD_TYPE_PREFIX",    kDSStdRecordTypePrefix,    CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_NATIVE_RECORD_TYPE_PREFIX", kDSNativeRecordTypePrefix, CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_STD_ATTR_TYPE_PREFIX",      kDSStdAttrTypePrefix,      CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_NATIVE_ATTR_TYPE_PREFIX",   kDSNativeAttrTypePrefix,   CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_ATTR_NONE",                 kDSAttrNone,               CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_STD_AUTH_METHOD_PREFIX",    kDSStdAuthMethodPrefix,    CONST_PERSISTENT | CONST_CS);
    REGISTER_STRING_CONSTANT("DS_NATIVE_AUTH_METHOD_PREFIX", kDSNativeAuthMethodPrefix, CONST_PERSISTENT | CONST_CS);
    
    return SUCCESS;
}
/* }}} */


ZEND_MSHUTDOWN_FUNCTION(opendirectory)
{
    UNREGISTER_INI_ENTRIES();
    return SUCCESS;
} 


ZEND_RINIT_FUNCTION(opendirectory)
{
    DS_G(separator) = estrdup("/");
    DS_G(status)    = eDSNoErr;
    return SUCCESS;
}

ZEND_RSHUTDOWN_FUNCTION(opendirectory)
{
    efree(DS_G(separator));
    return SUCCESS;
}


ZEND_MINFO_FUNCTION(opendirectory)
{
    char buff[32];
    
	php_info_print_table_start();
	php_info_print_table_header(2, "OpenDirectory Support", "enabled");
#if defined(DS_BUILDVERSION)
	php_info_print_table_row(2, "BuildVersion", DS_BUILDVERSION);
#endif
#if defined(DS_CFBUNDLEVERSION)
	php_info_print_table_row(2, "CFBundleVersion", DS_CFBUNDLEVERSION);
#endif
#if defined(DS_RELEASESTATUS)
	php_info_print_table_row(2, "ReleaseStatus", DS_RELEASESTATUS);
#endif
#if defined(DS_SOURCEVERSION)
	php_info_print_table_row(2, "SourceVersion", DS_SOURCEVERSION);
#endif
    sprintf(buff, "%ld", DS_G(num_refs));
    php_info_print_table_row(2, "Active OpenDirectory References", buff);
	php_info_print_table_end();
    
    DISPLAY_INI_ENTRIES();
}


#if COMPILE_DL_DIRECTORY_SERVICES
ZEND_GET_MODULE(opendirectory)
#endif



/* {{{
   }}} resource destructors */
static void _php_le_dirref_close(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
    php_t_dir_reference *dirref = (php_t_dir_reference *)rsrc->ptr;
    PHP_DBUFFER_DEALLOC(dirref);
    if (dirref->ref) {
        dsCloseDirService(dirref->ref);
    }
    DS_G(num_refs)--;
    efree(dirref);
}


static void _php_le_node_ref_close(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
    php_t_dir_node_ref *node_ref = (php_t_dir_node_ref *)rsrc->ptr;
    if (node_ref->ref) {
        dsCloseDirNode(node_ref->ref);
    }
    efree(node_ref);
}


static void _php_le_rec_ref_close(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
    php_t_rec_ref *rec_ref = (php_t_rec_ref *)rsrc->ptr;
    if (rec_ref->ref) {
        dsCloseRecord(rec_ref->ref);
    }
    efree(rec_ref);
}


/* {{{
   }}} private functions */
   
   
/* {{{ php_node_list_to_zarray 
 */
static tDirStatus php_node_list_to_zarray(php_t_dir_reference *dirref, php_t_data_buffer *dbuffer, zval *zarray TSRMLS_DC)
{
    zval         *z_path    = NULL;
    tDataListPtr node_name  = NULL;
    tDirStatus   dir_status = eDSNoErr;
    int          i;
    char         *path;

    for (i = 0; i < dbuffer->count; i++) {
        node_name = dsDataListAllocate(dirref->ref);
        if (node_name) {
            if ((dir_status = dsGetDirNodeName(dirref->ref, dbuffer->buffer, i+1, &node_name)) == eDSNoErr) {
                path = dsGetPathFromList(dirref->ref, node_name, DS_G(separator));
                MAKE_STD_ZVAL(z_path);
                ZVAL_STRINGL(z_path, path, strlen(path), 1);
                if(zend_hash_index_update(zarray->value.ht, i, (void *)&z_path, sizeof(zval *), NULL) == FAILURE) {
                    dir_status = dsDataListDeallocate(dirref->ref, node_name);
                    break;
                }
            } else {
                dir_status = dsDataListDeallocate(dirref->ref, node_name);
                break;
            }
        }
        dir_status = dsDataListDeallocate(dirref->ref, node_name);
    }
    dsDataListDeallocate(dirref->ref, node_name);
    
    return dir_status;
} 
/* }}} */


/* {{{ php_rec_list_to_zarray 
 */
static tDirStatus php_rec_list_to_zarray(php_t_dir_reference *dirref, php_t_data_buffer *dbuffer,
                                   php_t_dir_node_ref *node, zval *zarray, unsigned long rec_count TSRMLS_DC)
{
    zval               *z_rec_info_array  = NULL;
    tDirStatus         dir_status         = eDSNoErr;
    tAttributeListRef  attr_list_ref      = NULL;
    tRecordEntry       *rec_entry         = NULL;
    char               *type              = NULL, *name = NULL;
    int                i;

    for (i = 1; i <= rec_count && dir_status == eDSNoErr; i++) {
        if ((dir_status = dsGetRecordEntry(node->ref, dbuffer->buffer, i, &attr_list_ref, &rec_entry)) == eDSNoErr) {
        
            if ((dir_status = dsGetRecordNameFromEntry(rec_entry, &name)) != eDSNoErr) break;
            if ((dir_status = dsGetRecordTypeFromEntry(rec_entry, &type)) != eDSNoErr) break;
            
            MAKE_STD_ZVAL(z_rec_info_array);
            array_init(z_rec_info_array);
            add_assoc_long(z_rec_info_array,   DS_KEY_REC_ATTRCOUNT, rec_entry->fRecordAttributeCount);
            add_assoc_string(z_rec_info_array, DS_KEY_REC_NAME,      name, 1);
            add_assoc_string(z_rec_info_array, DS_KEY_REC_TYPE,      type, 1);
            
            zend_hash_index_update(zarray->value.ht, i-1, (void *)&z_rec_info_array, sizeof(zval *), NULL);
        } else {
            break;
        }
        dir_status = dsCloseAttributeList(attr_list_ref);
        dir_status = dsDeallocRecordEntry(dirref->ref, rec_entry);
        attr_list_ref = NULL;
        rec_entry     = NULL;
    }
  
    return dir_status;
} 
/* }}} */


/* {{{ php_zarray_to_data_list 
 */
static tDirStatus php_zarray_to_data_list(php_t_dir_reference *dirref, zval *zarray, tDataListPtr data_list TSRMLS_DC)
{
    zval          **entry = NULL;
    tDirStatus    dir_status = eDSNoErr;
    HashPosition  pos;
    
    zend_hash_internal_pointer_reset_ex(Z_ARRVAL_P(zarray), &pos);
    while (dir_status == eDSNoErr &&
            zend_hash_get_current_data_ex(Z_ARRVAL_P(zarray), (void **)&entry, &pos) == SUCCESS) {
        if (Z_TYPE_PP(entry) != IS_STRING) {
            convert_to_string_ex(entry);
        }
        dir_status = dsAppendStringToListAlloc(dirref->ref, data_list, Z_STRVAL_PP(entry));
        zend_hash_move_forward_ex(Z_ARRVAL_P(zarray), &pos);
    }

    return dir_status;
} 
/* }}} */


/* {{{ php_attr_list_to_zarray 
 */
static tDirStatus php_attr_list_to_zarray(php_t_dir_reference *dirref, php_t_dir_node_ref *node, 
                                          tAttributeListRef *attr_list, tAttributeValueListRef *attr_values,
                                          zval *zarray, long num_types, bool info_only TSRMLS_DC)
{
    zval                   *z_array_element = NULL;
    tDirStatus             dir_status       = eDSNoErr;
    tAttributeEntry        *attr_entry      = NULL;
    tAttributeValueEntry   *value_entry     = NULL;
    int i, k;

    for (i = 1; i <= num_types && dir_status == eDSNoErr; i++) 
    {
        if ((dir_status = dsGetAttributeEntry(node->ref, dirref->data->buffer, *attr_list, i, attr_values, &attr_entry)) == eDSNoErr) {
            if (!info_only) {
                MAKE_STD_ZVAL(z_array_element);
                array_init(z_array_element);
                for (k = 1; k <= attr_entry->fAttributeValueCount && dir_status == eDSNoErr; k++) {
                    if ((dir_status = dsGetAttributeValue(node->ref, dirref->data->buffer, k, *attr_values, &value_entry)) == eDSNoErr) {
                        
                        add_assoc_stringl(z_array_element, DS_KEY_ATTRVAL, value_entry->fAttributeValueData.fBufferData, 
                            value_entry->fAttributeValueData.fBufferLength+1, 1);
                        add_assoc_long(z_array_element, DS_KEY_ATTRVAL_ID, value_entry->fAttributeValueID);
                        dir_status  = dsDeallocAttributeValueEntry(dirref->ref, value_entry);
                        value_entry = NULL;
                        
                    } else {
                        efree(z_array_element);
                        break;
                    }
                }
                zend_hash_update(zarray->value.ht, attr_entry->fAttributeSignature.fBufferData, 
                    attr_entry->fAttributeSignature.fBufferLength+1, (void *)&z_array_element, sizeof(zval *), NULL);
            } else {
                add_index_stringl(zarray, i-1, attr_entry->fAttributeSignature.fBufferData, 
                    attr_entry->fAttributeSignature.fBufferLength+1, 1);
            } 
        }
    }
	dsDeallocAttributeEntry(dirref->ref, attr_entry);
    attr_entry = NULL;

    return dir_status;
} 
/* }}} */


/* {{{ php_alloc_node_auth_buffer 
 */
static php_node_auth_buffer *php_alloc_node_auth_buffer(php_t_dir_reference *dirref, char *user, 
                                                        char *passwd, int uid, char *auth_type_name TSRMLS_DC)
{
    php_node_auth_buffer *buffer = NULL;
    int  c = 0, t = 0, bufflen = 0;
    int  user_len = strlen(user), passwd_len = strlen(passwd);
    
    buffer = (php_node_auth_buffer *) emalloc(sizeof(php_node_auth_buffer));
    
	bufflen += sizeof(long) + user_len;
	bufflen += sizeof(long) + passwd_len;
	
    buffer->type   = dsDataNodeAllocateString(dirref->ref, auth_type_name);
    buffer->data   = dsDataBufferAllocate(dirref->ref, bufflen);
    buffer->result = dsDataBufferAllocate(dirref->ref, 512);

    /* put all of the authentication arguments into the buffer. */
    t = user_len;
    memcpy(&buffer->data->fBufferData[c], &t, sizeof(long));
    c += sizeof(long);
    memcpy(&buffer->data->fBufferData[c], user, t);
    c += t;
    
    t = passwd_len;
    memcpy(&buffer->data->fBufferData[c], &t, sizeof(long));
    c += sizeof(long);
    memcpy(&buffer->data->fBufferData[c], passwd, t);
    c += t;
    
    buffer->data->fBufferLength = bufflen;
    buffer->context = NULL;
    
    return buffer;
} 
/* }}} */


/* {{{ php_dealloc_node_auth_buffer 
 */
static void php_dealloc_node_auth_buffer(php_t_dir_reference *dirref, php_node_auth_buffer *buffer TSRMLS_DC)
{
    if (buffer->context) {
        dsReleaseContinueData(dirref->ref, buffer->context);
    }
    dsDataBufferDeAllocate(dirref->ref, buffer->data);
    dsDataBufferDeAllocate(dirref->ref, buffer->result);
    dsDataNodeDeAllocate(dirref->ref,   buffer->type);

    efree(buffer);
    buffer = NULL;
} 
/* }}} */


/* {{{ php_ds_open_directory_service 
 */
static void php_ds_open_directory_service(INTERNAL_FUNCTION_PARAMETERS, int mode)
{
    tDirStatus           dir_status   = eDSNoErr;
    php_t_dir_reference  *dirref      = NULL;
    php_node_auth_buffer *auth_buffer = NULL;
    char                 *passwd      = NULL, *auth_type_name = NULL, *host = NULL, *user = NULL;
    int                  user_len, passwd_len, auth_type_name_len, host_len;
    long                 portno;

    if (DS_G(max_refs) != -1 && DS_G(max_refs) >= DS_G(num_refs)) {
        zend_error(E_WARNING, "Too many directory references open [%ld]\n", DS_G(num_refs));
        RETURN_FALSE;
    }
    
    dirref = (php_t_dir_reference *) emalloc(sizeof(php_t_dir_reference));
    
    switch (mode) {
        case LOCAL_SESSION: {
            if ((dir_status = dsOpenDirService(&dirref->ref)) != eDSNoErr) {
                efree(dirref);
                PHP_DS_STATUS_RETURN_FALSE(dir_status);
            }
        }
        break;
        case REMOTE_SESSION: {
            if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "slss|s", &host, &host_len, 
                    &portno, &user, &user_len, &passwd, &passwd_len, &auth_type_name, &auth_type_name_len) == FAILURE) {
                WRONG_PARAM_COUNT;
            }
            if (!auth_type_name) {
                auth_type_name = kDSStdAuthNodeNativeClearTextOK;
            }
            auth_buffer = php_alloc_node_auth_buffer(dirref, user, passwd, 0, auth_type_name);
            if ((dir_status = dsOpenDirServiceProxy(&dirref->ref, host, portno, auth_buffer->type, 
                    auth_buffer->data, auth_buffer->result, auth_buffer->context)) != eDSNoErr) {
                php_dealloc_node_auth_buffer(dirref, auth_buffer);
                efree(dirref);
                PHP_DS_STATUS_RETURN_FALSE(dir_status);
            }
            php_dealloc_node_auth_buffer(dirref, auth_buffer);
        }
        break;
    }
    
    dirref->data = NULL;
    DS_G(num_refs)++;
    PHP_DS_STATUS(dir_status);
    RETURN_RESOURCE(zend_list_insert(dirref, le_dirref));
}
/* }}} */


/* {{{
   }}} php directory services functions */
   

/* {{{ proto resource ds_open_dir_service(void)
   Opens an Open Directory session. */
ZEND_FUNCTION(ds_open_dir_service)
{
    php_ds_open_directory_service(INTERNAL_FUNCTION_PARAM_PASSTHRU, LOCAL_SESSION);
}
/* }}} */


/* {{{ proto resource ds_open_dir_service_proxy(string host, int portno, string username, string passwd [, string auth_type_name])
   Opens a remote Open Directory session. */
ZEND_FUNCTION(ds_open_dir_service_proxy)
{
    php_ds_open_directory_service(INTERNAL_FUNCTION_PARAM_PASSTHRU, REMOTE_SESSION);
}
/* }}} */


/* {{{ proto void ds_set_separator(string sep)
   Sets the directory separator. */
ZEND_FUNCTION(ds_set_separator)
{
    char *sep;
    int  seplen;
    if (zend_parse_parameters(1 TSRMLS_CC, "s", &sep, &seplen) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    efree(DS_G(separator));
    DS_G(separator) = estrdup(sep);
}
/* }}} */


/* {{{ proto void ds_get_separator(void)
   Gets the directory separator. */
ZEND_FUNCTION(ds_get_separator)
{
    if (ZEND_NUM_ARGS() != 0) WRONG_PARAM_COUNT; 
    RETURN_STRING(DS_G(separator), 1);
}
/* }}} */


/* {{{ proto bool ds_close_dir_service(resource dir_ref)
   Closes a directory service reference. */
ZEND_FUNCTION(ds_close_dir_service)
{
    zval                *z_dirref = NULL;
    php_t_dir_reference *dirref   = NULL;

    if (zend_parse_parameters(1 TSRMLS_CC, "r", &z_dirref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    PHP_DBUFFER_DEALLOC(dirref);
    PHP_DS_STATUS(dsCloseDirService(dirref->ref));
    zend_list_delete(Z_LVAL_P(z_dirref));
    
    RETURN_TRUE;
}
/* }}} */


/* {{{ proto int ds_verify_dir_ref_num(resource dir_ref)
   Verifies a Directory Services reference number is valid. */
ZEND_FUNCTION(ds_verify_dir_ref_num)
{
    zval                *z_dirref  = NULL;
    php_t_dir_reference *dirref    = NULL;

    if (zend_parse_parameters(1 TSRMLS_CC, "r", &z_dirref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    
    PHP_DS_STATUS(dsVerifyDirRefNum(dirref->ref));
    RETURN_LONG(DS_G(status));
}
/* }}} */


/* {{{ proto int ds_add_child_pid_to_reference(resource dir_ref, resource node_ref, int pid)
   Allows a child process that your application forks to use a directory node reference
   that the parent process has already acquired.*/
ZEND_FUNCTION(ds_add_child_pid_to_reference)
{
    zval                *z_dirref  = NULL, *z_noderef = NULL;
    php_t_dir_reference *dirref    = NULL;
    php_t_dir_node_ref  *node_ref  = NULL;
    long                child_pid;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rrl", &z_dirref, &node_ref, &child_pid) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    ZEND_FETCH_RESOURCE(node_ref, php_t_dir_node_ref *, &z_noderef, -1, DS_RSRC_NAME_NODEREF, le_node_ref);
    
    PHP_DS_STATUS(dsAddChildPIDToReference(dirref->ref, child_pid, node_ref->ref));
    RETURN_LONG(DS_G(status));
}
/* }}} */


/* {{{ proto int ds_get_dir_node_count(resource dir_ref)
   Gets a directory's node count. */
ZEND_FUNCTION(ds_get_dir_node_count)
{
    zval                *z_dirref  = NULL;
    unsigned long       node_count = 0;
    php_t_dir_reference *dirref    = NULL;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &z_dirref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    PHP_DS_STATUS(dsGetDirNodeCount(dirref->ref, &node_count));
    
    RETURN_LONG(node_count);
}
/* }}} */


/* {{{ proto array ds_get_dir_node_list(resource dir_ref)
   Gets the names of registered nodes. */
ZEND_FUNCTION(ds_get_dir_node_list)
{
    zval                *z_dirref  = NULL; 
    php_t_dir_reference *dirref    = NULL;
    tDirStatus          dir_status = eDSNoErr;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &z_dirref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    
    if (dirref->data && dirref->data->buffer) {
        array_init(return_value);
        dir_status = dsGetDirNodeList(dirref->ref, dirref->data->buffer, &dirref->data->count, &dirref->data->context);
        dir_status = php_node_list_to_zarray(dirref, dirref->data, return_value TSRMLS_CC);
    } else {
        zend_error(E_WARNING, "no data buffer allocated: %s() should be called after a call to ds_data_buffer_alloc()\n", 
                get_active_function_name(TSRMLS_C));
    }
    PHP_DS_STATUS(dir_status);
}
/* }}} */


/* {{{ proto int ds_dir_status(resource dir_ref)
   Returns the status code from the last directory service function call. */
ZEND_FUNCTION(ds_dir_status)
{
    if (ZEND_NUM_ARGS() != 0) WRONG_PARAM_COUNT; 
    RETURN_LONG(DS_G(status));
}
/* }}} */


/* {{{ proto string ds_dir_errstr(int dir_status)
   Returns the error string corresponding to a directory service error code. */
ZEND_FUNCTION(ds_dir_errstr)
{
    long dir_status = eDSNoErr;
    
    if (zend_parse_parameters(1 TSRMLS_CC, "l", &dir_status) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    RETURN_STRING(php_dir_status_errstr(dir_status), 1);
}
/* }}} */


/* {{{ proto bool ds_is_dir_service_running(void)
   Checks if the directory service is running. */
ZEND_FUNCTION(ds_is_dir_service_running)
{
    if (ZEND_NUM_ARGS() != 0) WRONG_PARAM_COUNT;
    RETURN_BOOL(dsIsDirServiceRunning() == eDSNoErr ? 1 : 0);
}
/* }}} */


/* {{{ proto bool ds_data_buffer_alloc(resource dir_ref, int buffsize)
   Allocates memeory to the data buffer corresponding to a directory service reference. */
ZEND_FUNCTION(ds_data_buffer_alloc)
{
    zval                *z_dirref    = NULL;
    zend_bool           isbuffer     = 0;
    unsigned long       buffsize     = 0;
    php_t_dir_reference *dirref      = NULL;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", &z_dirref, &buffsize) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    
    if (dirref->data) {
        if (dirref->data->buffer) {
            dsDataBufferDeAllocate(dirref->ref, dirref->data->buffer);
            dirref->data->buffer = dsDataBufferAllocate(dirref->ref, buffsize);
            if (dirref->data->buffer) isbuffer = 1;
        }
    } else {
        PHP_DBUFFER_P_INIT(dirref);
        if ((dirref->data->buffer = dsDataBufferAllocate(dirref->ref, buffsize)) != NULL) isbuffer = 1;
    }

    RETURN_BOOL(isbuffer);
}
/* }}} */


/* {{{ proto bool ds_data_buffer_dealloc(resource dir_ref)
   Deallocates memeory to the data buffer corresponding to a directory service reference. */
ZEND_FUNCTION(ds_data_buffer_dealloc)
{
    zval                *z_dirref = NULL;
    php_t_dir_reference *dirref   = NULL;

    if (zend_parse_parameters(1 TSRMLS_CC, "r", &z_dirref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    
    PHP_DBUFFER_DEALLOC(dirref);
    RETURN_BOOL(dirref->status == eDSNoErr ? 1 : 0);
}
/* }}} */


/* {{{ proto bool ds_do_dir_node_auth(resource node, string auth_type, string user, string passwd [, bool auth_only])
   Authenticates a user to a directory node. */
ZEND_FUNCTION(ds_do_dir_node_auth)
{
    zval                 *z_noderef    = NULL;
    php_t_dir_reference  *dirref       = NULL;
    php_t_dir_node_ref   *node_ref     = NULL;
    php_node_auth_buffer *auth_buffer  = NULL;
    tDirStatus           dir_status    = eDSNoErr;
    char                 *user         = NULL, *passwd = NULL, *auth_type_name = NULL;
    zend_bool            auth_only = 0, result = 1;
    int                  user_len, passwd_len, auth_type_name_len, uid;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsss|b", 
            &z_noderef, &auth_type_name, &auth_type_name_len, &user, &user_len, &passwd, &passwd_len, &auth_only) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(node_ref, php_t_dir_node_ref *, &z_noderef, -1, DS_RSRC_NAME_NODEREF, le_node_ref);
    dirref = zend_list_find(node_ref->assoc_dirref_id, &le_dirref);

    auth_buffer = php_alloc_node_auth_buffer(dirref, user, passwd, uid, auth_type_name TSRMLS_CC);
    if ((dir_status = dsDoDirNodeAuth(node_ref->ref, auth_buffer->type, auth_only, auth_buffer->data, 
            auth_buffer->result, &auth_buffer->context)) != eDSNoErr) {
        result = 0;
    }
    
    php_dealloc_node_auth_buffer(dirref, auth_buffer TSRMLS_CC);
    PHP_DS_STATUS(dir_status);
    RETURN_BOOL(result);
}
/* }}} */


/* {{{ proto array ds_find_dir_nodes(resource dir_ref, string path [, int match_type])
    Finds directory nodes. */
ZEND_FUNCTION(ds_find_dir_nodes)
{
    zval                *z_dirref  = NULL; 
    tDirStatus          dir_status = eDSNoErr;
    tDataListPtr        node_name  = NULL;
    php_t_dir_reference *dirref    = NULL;
    tDirPatternMatch    match_type = eDSContains;
    char                *path;
    int                 pathlen;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs|l", &z_dirref, &path, &pathlen, &match_type) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);

    if (dirref->data && dirref->data->buffer) {
        node_name = dsBuildFromPath(dirref->ref, path, DS_G(separator));
        if (node_name) {
            array_init(return_value);
            dir_status = dsFindDirNodes(dirref->ref, dirref->data->buffer, node_name, match_type, &dirref->data->count, &dirref->data->context);
            dir_status = php_node_list_to_zarray(dirref, dirref->data, return_value TSRMLS_CC);
        }
        dsDataListDeallocate(dirref->ref, node_name);
    } else {
        zend_error(E_WARNING, "no data buffer allocated: %s() should be called after a call to ds_data_buffer_alloc()\n", 
                get_active_function_name(TSRMLS_C));
    }
    PHP_DS_STATUS(dir_status);
}
/* }}} */


/* {{{ proto resource ds_open_dir_node(resource dir_ref, string path)
   Opens a directory node. */
ZEND_FUNCTION(ds_open_dir_node)
{
    zval                *z_dirref  = NULL;
    tDirStatus          dir_status = eDSNoErr;
    php_t_dir_reference *dirref    = NULL;
    tDataListPtr        node_path  = NULL;
    php_t_dir_node_ref  *node_ref  = NULL;
    char                *path      = NULL;
    int                 pathlen;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &z_dirref, &path, &pathlen) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    
    node_path = dsBuildFromPath(dirref->ref, path, DS_G(separator));
    node_ref = (php_t_dir_node_ref *) emalloc(sizeof(php_t_dir_node_ref)+4);
    if (node_path) {
    	if ((dir_status = dsOpenDirNode(dirref->ref, node_path, &node_ref->ref)) == eDSNoErr) {
            node_ref->assoc_dirref_id = z_dirref->value.lval;
            PHP_DS_STATUS(dir_status);
            RETURN_RESOURCE(zend_list_insert(node_ref, le_node_ref));
        }
    }
    
    PHP_DS_STATUS_RETURN_FALSE(dir_status);
}
/* }}} */


/* {{{ proto bool ds_close_dir_node(resource node_ref)
   Closes a directory node. */
ZEND_FUNCTION(ds_close_dir_node)
{
    zval                *z_node_ref = NULL;
    php_t_dir_reference *dirref     = NULL;
    php_t_dir_node_ref  *node_ref   = NULL;
    tDirStatus          dir_status  = eDSNoErr;
    
    if (zend_parse_parameters(1 TSRMLS_CC, "r", &z_node_ref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(node_ref, php_t_dir_node_ref *, &z_node_ref, -1, DS_RSRC_NAME_NODEREF, le_node_ref);

    dirref = zend_list_find(node_ref->assoc_dirref_id, &le_dirref);
    if ((dir_status = dsCloseDirNode(node_ref->ref)) == eDSNoErr) {
        node_ref->ref = NULL;
        zend_list_delete(Z_LVAL_P(z_node_ref));
        PHP_DS_STATUS_RETURN_TRUE(dir_status);
    }
    
    PHP_DS_STATUS_RETURN_FALSE(dir_status);
}
/* }}} */


/* {{{ proto array ds_get_record_list(resource dir_ref, resource node_ref [, mixed rec_types [, mixed rec_names [, mixed attr_types]]])
   Get a list of records form a directory node. */
ZEND_FUNCTION(ds_get_record_list)
{
    zval                 *z_dirref  = NULL, *z_node_ref = NULL, *z_rec_types = NULL, *z_rec_names = NULL, *z_attr_types = NULL;
    tDirStatus           dir_status = eDSNoErr;
    php_t_dir_reference  *dirref    = NULL;
    php_t_dir_node_ref   *node_ref  = NULL;
    unsigned long        rec_count  = 0;
    tDataList            rec_names, rec_types, attr_types;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rr|zzz", &z_dirref, &z_node_ref, &z_rec_types, &z_rec_names, &z_attr_types) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    ZEND_FETCH_RESOURCE(node_ref, php_t_dir_node_ref *, &z_node_ref, -1, DS_RSRC_NAME_NODEREF, le_node_ref);
    
    if (dirref->data && dirref->data->buffer) {
        /* reverse fall through */
        switch (ZEND_NUM_ARGS()) {
            case 5:
                if (Z_TYPE_P(z_attr_types) == IS_ARRAY) {
                    dir_status = php_zarray_to_data_list(dirref, z_attr_types, &attr_types TSRMLS_CC);
                } else {
                    convert_to_string(z_attr_types);
                    dir_status = dsBuildListFromStringsAlloc(dirref->ref, &attr_types, Z_STRVAL_P(z_attr_types), NULL);
                }
             case 4:
                if (Z_TYPE_P(z_rec_names) == IS_ARRAY) {
                    dir_status = php_zarray_to_data_list(dirref, z_rec_names, &attr_types TSRMLS_CC);
                } else {
                    convert_to_string(z_rec_names);
                    dir_status = dsBuildListFromStringsAlloc(dirref->ref, &rec_names, Z_STRVAL_P(z_rec_names), NULL);
                }
             case 3:
                if (Z_TYPE_P(z_rec_types) == IS_ARRAY) {
                    dir_status = php_zarray_to_data_list(dirref, z_rec_types, &attr_types TSRMLS_CC);
                } else {
                    convert_to_string(z_rec_types);
                    dir_status = dsBuildListFromStringsAlloc(dirref->ref, &rec_types, Z_STRVAL_P(z_rec_types), NULL);
                }
                break;
            default:
                dir_status = dsBuildListFromStringsAlloc(dirref->ref, &rec_types,  kDSStdRecordTypeUsers, NULL);
                dir_status = dsBuildListFromStringsAlloc(dirref->ref, &rec_names,  kDSRecordsAll,         NULL);
                dir_status = dsBuildListFromStringsAlloc(dirref->ref, &attr_types, kDSAttributesAll,      NULL);
        }

        dir_status = dsGetRecordList(node_ref->ref, dirref->data->buffer, &rec_names, eDSExact, &rec_types, &attr_types, false, &rec_count, &dirref->data->context);
        array_init(return_value);
        dir_status = php_rec_list_to_zarray(dirref, dirref->data, node_ref, return_value, rec_count TSRMLS_CC);
        
        dsDataListDeallocate(dirref->ref, &rec_names);
        dsDataListDeallocate(dirref->ref, &rec_types);
        dsDataListDeallocate(dirref->ref, &attr_types);
    } else {
        zend_error(E_WARNING, "no data buffer allocated: %s() should be called after a call to ds_data_buffer_alloc()\n", 
            get_active_function_name(TSRMLS_C));
    }
    PHP_DS_STATUS(dir_status);
    if (dir_status != eDSNoErr) RETURN_NULL();
}
/* }}} */


/* {{{ proto bool ds_continue_data(resource dir_ref)
   Checks if there is more context data to be got in the buffer associated with a directory service reference. */
ZEND_FUNCTION(ds_continue_data)
{
    zval                *z_dirref = NULL; 
    php_t_dir_reference *dirref    = NULL;
    
    if (zend_parse_parameters(1 TSRMLS_CC, "r", &z_dirref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    RETURN_BOOL(dirref->data && dirref->data->context ? 1 : 0);
}
/* }}} */


/* {{{ proto bool ds_release_continue_data(resource dir_ref)
   Releases the memory used by the context data from the buffer associated with a directory service reference. */
ZEND_FUNCTION(ds_release_continue_data)
{
    zval                *z_dirref = NULL; 
    php_t_dir_reference *dirref   = NULL;
    
    if (zend_parse_parameters(1 TSRMLS_CC, "r", &z_dirref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    if (dirref->data) {
        if (dirref->data->context) {
            PHP_DS_STATUS(dsReleaseContinueData(dirref->ref, dirref->data->context));
            dirref->data->context = NULL;
            RETURN_TRUE;
        }
    }

    RETURN_FALSE;
}
/* }}} */


/* {{{ proto resource ds_open_record(resource node_ref, string node_name, string node_type [, int flags])
   Opens a record from a directory node. */
ZEND_FUNCTION(ds_open_record)
{ 
    zval                *z_node_ref = NULL;
    tDirStatus          dir_status  = eDSNoErr, flag;
    php_t_dir_node_ref  *node_ref   = NULL;
    php_t_dir_reference *dirref     = NULL;
    php_t_rec_ref       *rec_ref    = NULL;
    char                *name = NULL, *type = NULL;
    tDataNodePtr        rec_name = NULL, rec_type = NULL;
    int                 rec_name_len, rec_type_len;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss|l", &z_node_ref, &name, &rec_name_len, &type, &rec_type_len, &flag) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(node_ref, php_t_dir_node_ref *, &z_node_ref, -1, DS_RSRC_NAME_NODEREF, le_node_ref);
    
    dirref  = zend_list_find(node_ref->assoc_dirref_id, &le_dirref);
    rec_ref = (php_t_rec_ref *) emalloc(sizeof(php_t_rec_ref));
    rec_ref->assoc_dirref_id = node_ref->assoc_dirref_id;
    if ((rec_name = dsDataNodeAllocateString(dirref->ref, name)) != NULL) {
        if ((rec_type = dsDataNodeAllocateString(dirref->ref, type)) != NULL) {
            dir_status = dsOpenRecord(node_ref->ref, rec_type, rec_name, &rec_ref->ref);
            if (dir_status == eDSNoErr) {
                dsDataNodeDeAllocate(dirref->ref, rec_type);
                dsDataNodeDeAllocate(dirref->ref, rec_name);
                PHP_DS_STATUS(dir_status);
                RETURN_RESOURCE(zend_list_insert(rec_ref, le_rec_ref));
            }
        }
        dsDataNodeDeAllocate(dirref->ref, rec_type);
    }
    dsDataNodeDeAllocate(dirref->ref, rec_name);
    efree(rec_ref);
    
    PHP_DS_STATUS_RETURN_FALSE(dir_status);
}
/* }}} */


/* {{{ proto bool ds_close_record(resource rec_ref)
   Closes a directory node's record. */
ZEND_FUNCTION(ds_close_record)
{ 
    zval                *z_recref  = NULL;
    tDirStatus          dir_status = eDSNoErr;
    php_t_dir_reference *dirref    = NULL;
    php_t_rec_ref       *rec_ref   = NULL;

    if (zend_parse_parameters(1 TSRMLS_CC, "r", &z_recref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_recref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);
    dirref = zend_list_find(rec_ref->assoc_dirref_id, &le_dirref);
    
    if (rec_ref->ref) {
        if ((dir_status = dsCloseRecord(rec_ref->ref)) == eDSNoErr) {
            zend_list_delete(Z_LVAL_P(z_recref));
            PHP_DS_STATUS_RETURN_TRUE(dir_status);
        }
    }
    
    PHP_DS_STATUS_RETURN_FALSE(dir_status);
}
/* }}} */


/* {{{ proto int ds_flush_record(resource rec_ref)
   Writes a record. */
ZEND_FUNCTION(ds_flush_record)
{ 
    zval           *z_recref  = NULL;
    php_t_rec_ref  *rec_ref   = NULL;

    if (zend_parse_parameters(1 TSRMLS_CC, "r", &z_recref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_recref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);
    
    PHP_DS_STATUS(dsFlushRecord(rec_ref->ref));
    RETURN_LONG(DS_G(status));
}
/* }}} */


/* {{{ proto bool ds_set_record_name(resource rec_ref, string name)
   Sets a records name. */
ZEND_FUNCTION(ds_set_record_name)
{
    zval                *z_node_ref  = NULL, *z_rec_ref = NULL;
    zend_bool           done         = 0;
    tDirStatus          dir_status   = eDSNoErr;
    php_t_dir_reference *dirref      = NULL;
    php_t_rec_ref       *rec_ref     = NULL;
    tDataNodePtr        rec_name     = NULL;
    char                *new_name    = NULL;
    int                 new_name_len;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rrs|z", &z_node_ref, &z_rec_ref, &new_name, &new_name_len) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);
    
    dirref = zend_list_find(rec_ref->assoc_dirref_id, &le_dirref);
    if ((rec_name = dsDataNodeAllocateString(dirref->ref, new_name)) != NULL) {
        dir_status = dsSetRecordName(rec_ref->ref, rec_name);
        dsDataNodeDeAllocate(dirref->ref, rec_name);
        rec_name = NULL;
        done = dir_status == eDSNoErr ? 1 : 0;
    }
    
    PHP_DS_STATUS(dir_status);
    RETURN_BOOL(done);
}
/* }}} */


/* {{{ proto bool ds_set_record_type(resource rec_ref, string name)
   Sets a records type. */
ZEND_FUNCTION(ds_set_record_type)
{
    zval                *z_rec_ref = NULL;
    zend_bool           done         = 0;
    tDirStatus          dir_status   = eDSNoErr;
    php_t_dir_reference *dirref      = NULL;
    php_t_rec_ref       *rec_ref     = NULL;
    char                *new_type    = NULL;
    tDataNodePtr        new_rec_type = NULL;
    int                 new_type_len;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &z_rec_ref, &new_type, &new_type_len) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);
    
    dirref = zend_list_find(rec_ref->assoc_dirref_id, &le_dirref);
    if ((new_rec_type = dsDataNodeAllocateString(dirref->ref, new_type)) != NULL) {
        dir_status = dsSetRecordType(rec_ref->ref, new_rec_type);
        dsDataNodeDeAllocate(dirref->ref, new_rec_type);
        done = dir_status == eDSNoErr ? 1 : 0;
    }
    
    PHP_DS_STATUS(dir_status);
    RETURN_BOOL(done);
}
/* }}} */


/* {{{ proto mixed ds_create_record(resource dir_ref, resource node_ref, string name, string type)
   Create a new node record, on success returns <code>TRUE</code>, <code>FALSE</code> otherwisw. */
ZEND_FUNCTION(ds_create_record)
{
    zval                *z_dirref    = NULL, *z_node_ref  = NULL;
    tDirStatus          dir_status   = eDSNoErr;
    php_t_dir_node_ref  *node_ref    = NULL;
    php_t_dir_reference *dirref      = NULL;
    char                *name        = NULL, *type = NULL;
    tDataNodePtr        rec_name     = NULL, rec_type = NULL;
    int                 name_len, rec_type_len;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rrss", &z_dirref, &z_node_ref, &name, &name_len, &type, &rec_type_len) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(node_ref, php_t_dir_node_ref *, &z_node_ref, -1, DS_RSRC_NAME_NODEREF, le_node_ref);
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    
    rec_name = dsDataNodeAllocateString(dirref->ref, name);
    if (rec_name) {
        rec_type = dsDataNodeAllocateString(dirref->ref, type);
        if (rec_type) {
             if ((dir_status = dsCreateRecord(node_ref->ref, rec_type, rec_name)) == eDSNoErr) {
                PHP_DS_STATUS(dir_status);
                dsDataNodeDeAllocate(dirref->ref, rec_name);
                dsDataNodeDeAllocate(dirref->ref, rec_type);
                RETURN_TRUE;
            }
        }
        dsDataNodeDeAllocate(dirref->ref, rec_type);
    }
    dsDataNodeDeAllocate(dirref->ref, rec_name);
    PHP_DS_STATUS_RETURN_FALSE(dir_status);
}
/* }}} */


/* {{{ proto mixed ds_create_record_and_open(resource dir_ref, resource node_ref, string name, string type)
   Create a new node record and opens it. On success returns a record reference resource, or <code>FALSE</code> otherwise. */
ZEND_FUNCTION(ds_create_record_and_open)
{
    zval                *z_dirref    = NULL, *z_node_ref  = NULL;
    tDataNodePtr        rec_name     = NULL, rec_type = NULL;
    tDirStatus          dir_status   = eDSNoErr;
    php_t_dir_node_ref  *node_ref    = NULL;
    php_t_dir_reference *dirref      = NULL;
    php_t_rec_ref       *rec_ref     = NULL;
    char                *name        = NULL, *type = NULL;
    int                 name_len, type_len;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rrss", &z_dirref, &z_node_ref, &name, &name_len, &type, &type_len) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(node_ref, php_t_dir_node_ref *, &z_node_ref, -1, DS_RSRC_NAME_NODEREF, le_node_ref);
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    
    if ((rec_name = dsDataNodeAllocateString(dirref->ref, name)) != NULL) {
        if ((rec_type = dsDataNodeAllocateString(dirref->ref, type)) != NULL) {
        
            rec_ref = (php_t_rec_ref *) emalloc(sizeof(php_t_rec_ref));
            rec_ref->assoc_dirref_id = node_ref->assoc_dirref_id;
            if ((dir_status = dsCreateRecordAndOpen(node_ref->ref, rec_type, rec_name, &rec_ref->ref)) != eDSNoErr) {
                efree(rec_ref);
                PHP_DS_STATUS_RETURN_FALSE(dir_status);
            }
            
            PHP_DS_STATUS(dir_status);
            dsDataNodeDeAllocate(dirref->ref, rec_name);
            dsDataNodeDeAllocate(dirref->ref, rec_type);
            RETURN_RESOURCE(zend_list_insert(rec_ref, le_rec_ref));
        }
        dsDataNodeDeAllocate(dirref->ref, rec_type);
    }
    dsDataNodeDeAllocate(dirref->ref, rec_name);
    PHP_DS_STATUS_RETURN_FALSE(dir_status);
}
/* }}} */


/* {{{ proto bool ds_add_attribute(resource rec_ref, string type, string value)
   Adds an attribute to a record. */
ZEND_FUNCTION(ds_add_attribute)
{
    zval                *z_rec_ref  = NULL;
    tDataNodePtr        attr_type   = NULL, attr_val = NULL;
    tDirStatus          dir_status  = eDSNoErr;
    zend_bool           done        = 0;
    php_t_dir_reference *dirref     = NULL;
    php_t_rec_ref       *rec_ref    = NULL;
    char                *type       = NULL, *value = NULL;
    long                type_len, value_len;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss", &z_rec_ref, &type, &type_len, &value, &value_len) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);

    dirref = zend_list_find(rec_ref->assoc_dirref_id, &le_dirref);
    if (dirref->ref) {
        if ((attr_type = dsDataNodeAllocateString(dirref->ref, type)) != NULL) {
            if ((attr_val = dsDataNodeAllocateString(dirref->ref, value)) != NULL) {
                dir_status = dsAddAttribute(rec_ref->ref, attr_type, NULL, attr_val);
                if (dir_status == eDSNoErr) done = 1;
                dsDataNodeDeAllocate(dirref->ref, attr_val);
            }
            dsDataNodeDeAllocate(dirref->ref, attr_type);
        }
    }
    PHP_DS_STATUS(dir_status);
    
    RETURN_BOOL(done);
}
/* }}} */


/* {{{ proto bool ds_add_attribute_value(resource rec_ref, string type, mixed value)
   Adds a value to a record's attribute. */
ZEND_FUNCTION(ds_add_attribute_value)
{
    zval                **z_rec_ref  = NULL, **z_attr_type = NULL, **z_attr_value = NULL;
    tDataNodePtr        attr_name    = NULL, attr_val = NULL;
    tDirStatus          dir_status   = eDSNoErr;
    zend_bool           done         = 0;
    php_t_dir_reference *dirref      = NULL;
    php_t_rec_ref       *rec_ref     = NULL;

    if (zend_get_parameters_ex(3, &z_rec_ref, &z_attr_type, &z_attr_value) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);
    
    dirref = zend_list_find(rec_ref->assoc_dirref_id, &le_dirref);
    if (dirref->ref) {
        convert_to_string_ex(z_attr_type);
        attr_name = dsDataNodeAllocateString(dirref->ref, Z_STRVAL_PP(z_attr_type));
        if (attr_name) {
            switch (Z_TYPE_PP(z_attr_value)) {
                case IS_ARRAY: 
                    {
                        HashPosition  pos;
                        zval **entry = NULL;
                        zend_hash_internal_pointer_reset_ex(Z_ARRVAL_PP(z_attr_value), &pos);
                        while (dir_status == eDSNoErr &&
                                zend_hash_get_current_data_ex(Z_ARRVAL_PP(z_attr_value), (void **)&entry, &pos) == SUCCESS) {
                                
                            convert_to_string_ex(entry);
                            attr_val = dsDataNodeAllocateString(dirref->ref, Z_STRVAL_PP(entry));
                            dir_status = dsAddAttributeValue(rec_ref->ref, attr_name, attr_val);
                            
                            zend_hash_move_forward_ex(Z_ARRVAL_PP(z_attr_value), &pos);
                            dsDataNodeDeAllocate(dirref->ref, attr_val);
                            attr_val = NULL;
                        }
                    }
                    break;
                case IS_STRING:
                /* fall through */
                default:
                    convert_to_string_ex(z_attr_value);
                    attr_val = dsDataNodeAllocateString(dirref->ref, Z_STRVAL_PP(z_attr_value));
                    dir_status = dsAddAttributeValue(rec_ref->ref, attr_name, attr_val);
                    dsDataNodeDeAllocate(dirref->ref, attr_val);
                    break;
            }
            if (dir_status == eDSNoErr) {
                done = 1;
            }
        }
        dsDataNodeDeAllocate(dirref->ref, attr_name);
    }
    
    PHP_DS_STATUS(dir_status);
    RETURN_BOOL(done);
}
/* }}} */


/* {{{ proto bool ds_remove_attribute(resource rec_ref, string name)
   Removes an attribute and its value from a record. */
ZEND_FUNCTION(ds_remove_attribute)
{
    zval                *z_rec_ref  = NULL;
    tDirStatus          dir_status  = eDSNoErr;
    zend_bool           done        = 0;
    php_t_dir_reference *dirref     = NULL;
    php_t_rec_ref       *rec_ref    = NULL;
    tDataNodePtr        attr_name   = NULL;
    char                *name       = NULL;
    int                 name_len;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs|z", &z_rec_ref, &name, &name_len) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);

    dirref = zend_list_find(rec_ref->assoc_dirref_id, &le_dirref);
    attr_name = dsDataNodeAllocateString(dirref->ref, name);
    if (attr_name) {
        dir_status = dsRemoveAttribute(rec_ref->ref, attr_name);
        if (dir_status == eDSNoErr) done = 1;
    } 
    dsDataNodeDeAllocate(dirref->ref, attr_name);   
    PHP_DS_STATUS(dir_status);
    
    RETURN_BOOL(done);
}
/* }}} */


/* {{{ proto bool ds_remove_attribute_value(resource rec_ref, string type, int value_id)
   Removes an attribute value from a record's attribute. */
ZEND_FUNCTION(ds_remove_attribute_value)
{
    zval                *z_rec_ref = NULL;
    tDirStatus          dir_status   = eDSNoErr;
    zend_bool           done         = 0;
    php_t_dir_reference *dirref      = NULL;
    php_t_rec_ref       *rec_ref     = NULL;
    tDataNodePtr        attr_type    = NULL;
    char                *type        = NULL;
    long                type_len, attr_id;
 
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsl", &z_rec_ref, &type, &type_len, &attr_id) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);
    dirref = zend_list_find(rec_ref->assoc_dirref_id, &le_dirref);
    
    attr_type = dsDataNodeAllocateString(dirref->ref, type);
    if (attr_type) {
        dir_status = dsRemoveAttributeValue(rec_ref->ref, attr_type, attr_id);
        if (dir_status == eDSNoErr) done = 1;
    } 
    dsDataNodeDeAllocate(dirref->ref, attr_type);   
    PHP_DS_STATUS(dir_status);
    
    RETURN_BOOL(done);
}
/* }}} */


/* {{{ proto string ds_get_record_attr_value_by_index(resource rec_ref, string type, int index)
   Get the value specified at <code>index</code> of an attribute value list. */
ZEND_FUNCTION(ds_get_record_attr_value_by_index)
{
    zval                *z_rec_ref   = NULL;
    tDirStatus          dir_status   = eDSNoErr;
    php_t_dir_reference *dirref      = NULL;
    php_t_rec_ref       *rec_ref     = NULL;
    char                *type        = NULL;
    tDataNodePtr        attr_type    = NULL;
    tAttributeValueEntryPtr value_entry = NULL;
    int                 type_len;
    long                index;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsl", &z_rec_ref, &type, &type_len, &index) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);

    dirref = zend_list_find(rec_ref->assoc_dirref_id, &le_dirref);
            
    if ((attr_type = dsDataNodeAllocateString(dirref->ref, type)) != NULL) {
        dir_status = dsGetRecordAttributeValueByIndex(rec_ref->ref, attr_type, index, &value_entry);
        if (dir_status == eDSNoErr) {
            ZVAL_STRINGL(return_value, value_entry->fAttributeValueData.fBufferData, value_entry->fAttributeValueData.fBufferLength+1, 1);
            dir_status  = dsDeallocAttributeValueEntry(dirref->ref, value_entry);
            dsDataNodeDeAllocate(dirref->ref, attr_type);
            value_entry = NULL;
            return;
        }
    } 
    dsDataNodeDeAllocate(dirref->ref, attr_type);   
    PHP_DS_STATUS(dir_status);
    RETURN_EMPTY_STRING();
}
/* }}} */


/* {{{ proto string ds_get_record_attr_value_by_id(resource rec_ref, string type, int id)
   Uses an attribute value ID to obtain the value of an attribute. */
ZEND_FUNCTION(ds_get_record_attr_value_by_id)
{
    zval                *z_rec_ref   = NULL;
    tDirStatus          dir_status   = eDSNoErr;
    php_t_dir_reference *dirref      = NULL;
    php_t_rec_ref       *rec_ref     = NULL;
    char                *type        = NULL;
    tDataNodePtr        attr_type    = NULL;
    tAttributeValueEntryPtr value_entry = NULL;
    int                 type_len;
    long                id;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsl", &z_rec_ref, &type, &type_len, &index) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);

    dirref = zend_list_find(rec_ref->assoc_dirref_id, &le_dirref);
            
    if ((attr_type = dsDataNodeAllocateString(dirref->ref, type)) != NULL) {
        dir_status = dsGetRecordAttributeValueByID(rec_ref->ref, attr_type, id, &value_entry);
        if (dir_status == eDSNoErr) {
            ZVAL_STRINGL(return_value, value_entry->fAttributeValueData.fBufferData, value_entry->fAttributeValueData.fBufferLength+1, 1);
            dir_status  = dsDeallocAttributeValueEntry(dirref->ref, value_entry);
            dsDataNodeDeAllocate(dirref->ref, attr_type);
            value_entry = NULL;
            return;
        }
    } 
    dsDataNodeDeAllocate(dirref->ref, attr_type);   
    PHP_DS_STATUS(dir_status);
    RETURN_EMPTY_STRING();
}
/* }}} */


/* {{{ proto bool ds_delete_record(resource rec_ref)
   Deletes a record. */
ZEND_FUNCTION(ds_delete_record)
{ 
    zval           *z_rec_ref  = NULL;
    tDirStatus     dir_status  = eDSNoErr;
    php_t_rec_ref  *rec_ref    = NULL;

    if (zend_parse_parameters(1 TSRMLS_CC, "r", &z_rec_ref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);
    
    if (rec_ref->ref) {
        if ((dir_status = dsDeleteRecord(rec_ref->ref)) == eDSNoErr) {
            rec_ref->ref = NULL;
            zend_list_delete(Z_LVAL_P(z_rec_ref));
            convert_to_null(z_rec_ref);
            PHP_DS_STATUS_RETURN_TRUE(dir_status);
        }
    }
    
    PHP_DS_STATUS_RETURN_FALSE(dir_status);
}
/* }}} */


/* {{{ proto array ds_get_dir_node_info(resource node_ref [, array attr_type_strings [, bool attr_info_only]])
   Gets information about attribute types and attribute values for a directory node. */
ZEND_FUNCTION(ds_get_dir_node_info)
{ 
    zval                   *z_node_ref    = NULL, *z_attrs = NULL;
    zend_bool              attr_info_only = 0;
    tDirStatus             dir_status     = eDSNoErr;
    long                   num_types      = 0;
    php_t_dir_node_ref     *node_ref      = NULL;
    php_t_dir_reference    *dirref        = NULL;
    tAttributeListRef      attr_list      = NULL;
    tAttributeValueListRef attr_values    = NULL;
    tDataListPtr           attr_types     = NULL;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r|zb", &z_node_ref, &z_attrs, &attr_info_only) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(node_ref, php_t_dir_node_ref *, &z_node_ref, -1, DS_RSRC_NAME_NODEREF, le_node_ref);
    dirref = zend_list_find(node_ref->assoc_dirref_id, &le_dirref);
    
    if (!dirref->data) {
        zend_error(E_WARNING, "no data buffer allocated: %s() should be called after a call to ds_data_buffer_alloc()\n", 
                get_active_function_name(TSRMLS_C));
        RETURN_FALSE;
    }
    attr_types = dsDataListAllocate(dirref->ref);
    if (Z_TYPE_P(z_attrs) == IS_ARRAY) {
        dir_status = php_zarray_to_data_list(dirref, z_attrs, attr_types TSRMLS_CC);
    } else {
        convert_to_string(z_attrs);
        dir_status = dsBuildListFromStringsAlloc(dirref->ref, attr_types, Z_STRVAL_P(z_attrs), NULL);
    }
    dir_status = dsGetDirNodeInfo(node_ref->ref, attr_types, dirref->data->buffer, /* attr_info_only */ FALSE, &num_types, &attr_list, &dirref->data->context);
    if (dir_status == eDSNoErr) {
        array_init(return_value);
        dir_status = php_attr_list_to_zarray(dirref, node_ref, &attr_list, &attr_values, return_value, num_types, attr_info_only TSRMLS_CC);
    }
    dsCloseAttributeValueList(attr_values);
    dsCloseAttributeValueList(attr_list);
    dsDataListDeallocate(dirref->ref, attr_types);
    
    PHP_DS_STATUS(dir_status);
    if (dir_status != eDSNoErr) {
        zval_dtor(return_value);
        RETURN_FALSE;
    }
}
/* }}} */


/* {{{ proto array ds_get_record_reference_info(resource dirref, resource rec_ref)
   Gets information about a record. */
ZEND_FUNCTION(ds_get_record_reference_info)
{ 
    zval                 *z_dirref   = NULL, *z_rec_ref  = NULL;
    tDirStatus           dir_status  = eDSNoErr;
    php_t_rec_ref        *rec_ref    = NULL;
    php_t_dir_reference  *dirref     = NULL;
    tRecordEntry         *rec_entry  = NULL;

    if (zend_parse_parameters(2 TSRMLS_CC, "rr", &z_dirref, &z_rec_ref) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);
    
    if ((dir_status = dsGetRecordReferenceInfo(dirref->ref, &rec_entry)) == eDSNoErr) {
        array_init(return_value);
        add_assoc_long(return_value,    DS_KEY_REC_ATTRCOUNT, rec_entry->fRecordAttributeCount);
        add_assoc_stringl(return_value, DS_KEY_REC_NAMETYPE,  rec_entry->fRecordNameAndType.fBufferData, rec_entry->fRecordNameAndType.fBufferLength+1, 1);
    
        dsDeallocRecordEntry(dirref->ref, rec_entry);
        return;
    }
    
    PHP_DS_STATUS_RETURN_FALSE(dir_status);
}
/* }}} */


/* {{{ proto array ds_get_record_attribute_info(resource dirref, resource rec_ref, string attr_type)
   Gets information about a records attribute. */
ZEND_FUNCTION(ds_get_record_attribute_info)
{ 
    zval                 *z_dirref   = NULL, *z_rec_ref  = NULL;
    tDirStatus           dir_status  = eDSNoErr;
    php_t_rec_ref        *rec_ref    = NULL;
    php_t_dir_reference  *dirref     = NULL;
    tAttributeEntryPtr   attr        = NULL;
    tDataNodePtr         attr_type   = NULL;
    char                 *attr_type_name = NULL;
    long                 attr_type_name_len;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rrs", &z_dirref, &z_rec_ref, &attr_type_name , &attr_type_name_len) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(dirref, php_t_dir_reference *, &z_dirref, -1, DS_RSRC_NAME_DIRREF, le_dirref);
    ZEND_FETCH_RESOURCE(rec_ref, php_t_rec_ref *, &z_rec_ref, -1, DS_RSRC_NAME_RECREF, le_rec_ref);
    
    attr_type = dsDataNodeAllocateString(dirref->ref, attr_type_name);
    if ((dir_status = dsGetRecordAttributeInfo(rec_ref->ref, attr_type, &attr)) == eDSNoErr) {
        
        array_init(return_value);
        add_assoc_long(return_value,    DS_KEY_ATTRVAL_COUNT,   attr->fAttributeValueCount);
        add_assoc_long(return_value,    DS_KEY_ATTR_SIZE,       attr->fAttributeDataSize);
        add_assoc_long(return_value,    DS_KEY_ATTRVAL_MAXSIZE, attr->fAttributeValueMaxSize);
        add_assoc_stringl(return_value, DS_KEY_ATTR_SIG,        attr->fAttributeSignature.fBufferData, attr->fAttributeSignature.fBufferLength+1, 1);

        dsDataNodeDeAllocate(dirref->ref, attr_type);
        dsDeallocAttributeEntry(dirref->ref, attr);
        return;
    }
    
    dsDataNodeDeAllocate(dirref->ref, attr_type);
    PHP_DS_STATUS_RETURN_FALSE(dir_status);
}
/* }}} */


/* {{{ proto array ds_do_attribute_value_search(resource noderef, mixed rec_types, string attr_type, int match_type, string search)
   Searches a node for records by attribute value. */
ZEND_FUNCTION(ds_do_attribute_value_search)
{ 
    zval                   *z_node_ref    = NULL, *z_rec_types = NULL;
    tDirStatus             dir_status     = eDSNoErr;
    php_t_dir_node_ref     *node_ref      = NULL;
    php_t_dir_reference    *dirref        = NULL;
    tDataListPtr           rec_types      = NULL;
    tDataNodePtr           attr_type      = NULL;
    tDataNodePtr           searchdata     = NULL;
    char                   *search = NULL, *attr = NULL;
    long                   match_type, search_len, attr_len;
    unsigned long          num_matches = 0;
    
    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rzsls", 
        &z_node_ref, &z_rec_types, &attr, &attr_len, &match_type, &search, &search_len) == FAILURE) {
        WRONG_PARAM_COUNT;
    }
    ZEND_FETCH_RESOURCE(node_ref, php_t_dir_node_ref *, &z_node_ref, -1, DS_RSRC_NAME_NODEREF, le_node_ref);
    dirref = zend_list_find(node_ref->assoc_dirref_id, &le_dirref);
    
    if (!dirref->data) {
        zend_error(E_WARNING, "no data buffer allocated: %s() should be called after a call to ds_data_buffer_alloc()\n", 
            get_active_function_name(TSRMLS_C));
        RETURN_FALSE;
    }
    
    rec_types = dsDataListAllocate(dirref->ref);
    if (Z_TYPE_P(z_rec_types) == IS_ARRAY) {
        dir_status = php_zarray_to_data_list(dirref, z_rec_types, rec_types TSRMLS_CC);
    } else {
        convert_to_string(z_rec_types);
        dir_status = dsBuildListFromStringsAlloc(dirref->ref, rec_types, Z_STRVAL_P(z_rec_types), NULL);
    }
    
    attr_type  = dsDataNodeAllocateString(dirref->ref, attr);
    searchdata = dsDataNodeAllocateString(dirref->ref, search);
    dir_status = dsDoAttributeValueSearch(node_ref->ref, dirref->data->buffer, rec_types, attr_type, match_type,
                            searchdata, &num_matches, dirref->data->context);

    if (dir_status == eDSNoErr) {
        array_init(return_value);
        dir_status = php_rec_list_to_zarray(dirref, dirref->data, node_ref, return_value, num_matches TSRMLS_CC);
    }

    dsDataListDeallocate(dirref->ref, rec_types);
    dsDataNodeDeAllocate(dirref->ref, attr_type);
    dsDataNodeDeAllocate(dirref->ref, searchdata);
    
    PHP_DS_STATUS(dir_status);
    if (dir_status != eDSNoErr) {
        zval_dtor(return_value);
        RETURN_FALSE;
    }
}
/* }}} */


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: sw=4 ts=4 fdm=marker
 * vim<600: sw=4 ts=4
 */

